#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import argparse
import asyncio
import logging
from asyncio import CancelledError
from configparser import ConfigParser

from . import config, factory

logging.basicConfig(level=logging.WARNING)


async def build_and_start(app_config: ConfigParser):
    readers = factory.build_meters(app_config)
    sinks = factory.build_sinks(app_config)
    data_collector = factory.build_collector(readers, sinks)

    await asyncio.gather(*[sink.start() for sink in sinks])

    try:
        await asyncio.gather(
            *[reader.start() for reader in readers],
            data_collector.process_queue())
    except CancelledError:
        pass
    finally:
        logging.info("App shutting down now.")
        await asyncio.gather(*[sink.stop() for sink in sinks])


def set_logging_levels(app_config: ConfigParser) -> None:
    if not app_config.has_section("logging"):
        return
    # configure root logger
    logging.getLogger().setLevel(app_config["logging"].get('default', "WARNING"))
    # configure individual loggers
    for name, level in app_config["logging"].items():
        logging.getLogger(name).setLevel(level)


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Smart Meter Data Collector", add_help=True)
    parser.add_argument(
        '-c', '--config', help="File path of the configuration (.ini) file.", default="./datacollector.ini")
    parser.add_argument(
        '-s', '--saveconfig', help="Create default configuration (.ini) file at path defined with -c",
        action='store_true')
    parser.add_argument(
        '-d', '--dev', help="Development mode", action='store_true')
    return parser.parse_args()


def main():
    args = parse_arguments()
    debug_mode = bool(args.dev)
    if args.saveconfig:
        config.write_default_config(args.config)
        logging.warning("Default configuration written to file '%s'.", args.config)
        return
    app_config = config.read_config_files(args.config)
    set_logging_levels(app_config)
    asyncio.run(build_and_start(app_config), debug=debug_mode)



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from abc import ABC, abstractmethod
from typing import List, Optional

from .cosem import Cosem
from .hdlc_dlms_parser import HdlcDlmsParser
from .meter_data import MeterDataPoint
from .serial_reader import SerialConfig, SerialReader


class MeterError(Exception):
    pass


class Meter(ABC):
    def __init__(self) -> None:
        self._observers = []

    def register(self, observer) -> None:
        self._observers.append(observer)

    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError()

    def _notify_observers(self, data_points: List[MeterDataPoint]) -> None:
        for observer in self._observers:
            observer.notify(data_points)


class SerialHdlcDlmsMeter(Meter):
    HDLC_FLAG = b"\x7e"

    def __init__(self, serial_config: SerialConfig, cosem: Cosem, decryption_key: Optional[str] = None) -> None:
        super().__init__()
        self._parser = HdlcDlmsParser(cosem, decryption_key)
        self._serial = SerialReader(serial_config, self._data_received)

    async def start(self) -> None:
        await self._serial.start_and_listen()

    def _data_received(self, received_data: bytes) -> None:
        if not received_data:
            return
        if received_data == SerialHdlcDlmsMeter.HDLC_FLAG:
            self._parser.append_to_hdlc_buffer(received_data)
            return

        self._parser.append_to_hdlc_buffer(received_data)
        if self._parser.extract_data_from_hdlc_frames():
            dlms_objects = self._parser.parse_to_dlms_objects()
            data_points = self._parser.convert_dlms_bundle_to_reader_data(dlms_objects)
            self._notify_observers(data_points)



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from typing import Optional

import serial

from .cosem import Cosem, RegisterCosem
from .meter import MeterError, SerialHdlcDlmsMeter
from .meter_data import MeterDataPointTypes
from .reader import ReaderError
from .serial_reader import SerialConfig

LOGGER = logging.getLogger("smartmeter")

ISKRA_AM550_COSEM_REGISTERS = [
    RegisterCosem("1.1.1.7.0.255", MeterDataPointTypes.ACTIVE_POWER_P.value),
    RegisterCosem("1.1.2.7.0.255", MeterDataPointTypes.ACTIVE_POWER_N.value),
    RegisterCosem("1.1.3.7.0.255", MeterDataPointTypes.REACTIVE_POWER_P.value),
    RegisterCosem("1.1.4.7.0.255", MeterDataPointTypes.REACTIVE_POWER_N.value),

    RegisterCosem("1.0.13.7.0.255", MeterDataPointTypes.POWER_FACTOR.value, 0.001),
    RegisterCosem("1.0.14.7.0.255", MeterDataPointTypes.NET_FREQUENCY.value),

    RegisterCosem("1.0.21.7.0.255", MeterDataPointTypes.ACTIVE_POWER_P_L1.value),
    RegisterCosem("1.0.22.7.0.255", MeterDataPointTypes.ACTIVE_POWER_N_L1.value),
    RegisterCosem("1.0.23.7.0.255", MeterDataPointTypes.REACTIVE_POWER_P_L1.value),
    RegisterCosem("1.0.24.7.0.255", MeterDataPointTypes.REACTIVE_POWER_N_L1.value),

    RegisterCosem("1.0.31.7.0.255", MeterDataPointTypes.CURRENT_L1.value),
    RegisterCosem("1.0.32.7.0.255", MeterDataPointTypes.VOLTAGE_L1.value),

    RegisterCosem("1.0.41.7.0.255", MeterDataPointTypes.ACTIVE_POWER_P_L2.value),
    RegisterCosem("1.0.42.7.0.255", MeterDataPointTypes.ACTIVE_POWER_N_L2.value),
    RegisterCosem("1.0.43.7.0.255", MeterDataPointTypes.REACTIVE_POWER_P_L2.value),
    RegisterCosem("1.0.44.7.0.255", MeterDataPointTypes.REACTIVE_POWER_N_L2.value),

    RegisterCosem("1.0.51.7.0.255", MeterDataPointTypes.CURRENT_L2.value),
    RegisterCosem("1.0.52.7.0.255", MeterDataPointTypes.VOLTAGE_L2.value),

    RegisterCosem("1.0.61.7.0.255", MeterDataPointTypes.ACTIVE_POWER_P_L3.value),
    RegisterCosem("1.0.62.7.0.255", MeterDataPointTypes.ACTIVE_POWER_N_L3.value),
    RegisterCosem("1.0.63.7.0.255", MeterDataPointTypes.REACTIVE_POWER_P_L3.value),
    RegisterCosem("1.0.64.7.0.255", MeterDataPointTypes.REACTIVE_POWER_N_L3.value),

    RegisterCosem("1.0.71.7.0.255", MeterDataPointTypes.CURRENT_L3.value),
    RegisterCosem("1.0.72.7.0.255", MeterDataPointTypes.VOLTAGE_L3.value),

    RegisterCosem("1.0.81.7.40.255", MeterDataPointTypes.ANGLE_UI_L1.value),
    RegisterCosem("1.0.81.7.51.255", MeterDataPointTypes.ANGLE_UI_L2.value),
    RegisterCosem("1.0.81.7.62.255", MeterDataPointTypes.ANGLE_UI_L3.value),

    RegisterCosem("1.1.1.8.0.255", MeterDataPointTypes.ACTIVE_ENERGY_P.value),
    RegisterCosem("1.1.1.8.1.255", MeterDataPointTypes.ACTIVE_ENERGY_P_T1.value),
    RegisterCosem("1.1.1.8.2.255", MeterDataPointTypes.ACTIVE_ENERGY_P_T2.value),
    RegisterCosem("1.1.2.8.0.255", MeterDataPointTypes.ACTIVE_ENERGY_N.value),
    RegisterCosem("1.1.2.8.1.255", MeterDataPointTypes.ACTIVE_ENERGY_N_T1.value),
    RegisterCosem("1.1.2.8.2.255", MeterDataPointTypes.ACTIVE_ENERGY_N_T2.value),
    RegisterCosem("1.1.3.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_P.value),
    RegisterCosem("1.1.4.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_N.value),

    RegisterCosem("1.1.5.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_Q1.value),
    RegisterCosem("1.1.6.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_Q2.value),
    RegisterCosem("1.1.7.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_Q3.value),
    RegisterCosem("1.1.8.8.0.255", MeterDataPointTypes.REACTIVE_ENERGY_Q4.value),
]


class IskraAM550(SerialHdlcDlmsMeter):
    def __init__(self, port: str, decryption_key: Optional[str] = None) -> None:
        serial_config = SerialConfig(
            port=port,
            baudrate=115200,
            data_bits=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stop_bits=serial.STOPBITS_ONE,
            termination=IskraAM550.HDLC_FLAG
        )
        cosem = Cosem(
            fallback_id=port,
            register_obis=ISKRA_AM550_COSEM_REGISTERS
        )
        if decryption_key:
            LOGGER.warning("Using the Iskra AM550 meter with encrypted data has NOT BEEN TESTED yet!")
        try:
            super().__init__(serial_config, cosem, decryption_key)
        except ReaderError as ex:
            LOGGER.fatal("Unable to setup serial reader for Iskra AM550. '%s'", ex)
            raise MeterError("Failed setting up Iskra AM550.") from ex

        LOGGER.info("Successfully set up Iskra AM550 smart meter on '%s'.", port)



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional

from gurux_dlms import GXDateTime
from gurux_dlms.objects import GXDLMSClock, GXDLMSData

from .meter_data import MeterDataPointType

LOGGER = logging.getLogger("smartmeter")

OBIS_DEFAULT_IDS = [
    "0.0.42.0.0.255",
    "0.0.96.1.0.255",
    "0.0.96.2.0.255",
    "0.0.96.3.0.255",
    "0.0.96.4.0.255",
    "0.0.96.5.0.255"
]
OBIS_DEFAULT_CLOCK = "0.0.1.0.0.255"
COSEM_OBJECT_DETECT_ATTEMPTS = 3


@dataclass
class RegisterCosem:
    obis: str
    data_point_type: MeterDataPointType
    scaling: float = 1.0


class Cosem:
    def __init__(self, fallback_id: str, register_obis: List[RegisterCosem]) -> None:
        self._id: Optional[str] = None
        if not fallback_id:
            fallback_id = str(uuid.uuid1())
            LOGGER.warning("Empty or no fallback ID. Setting to random UUID %s.", fallback_id)
        self._fallback_id = fallback_id
        self._register_obis = {r.obis: r for r in register_obis}
        self._id_detect_countdown = COSEM_OBJECT_DETECT_ATTEMPTS

    def retrieve_id(self, dlms_objects: Dict[str, Any]) -> str:
        if self._id:
            return self._id

        id_obis = self._find_id_obis(dlms_objects)
        if not id_obis:
            LOGGER.debug("Unable to find ID object. Using fallback ID %s.", self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        id_obj = dlms_objects[id_obis]
        if not isinstance(id_obj, GXDLMSData):
            LOGGER.debug("Invalid ID object for OBIS code %s. Using fallback ID %s.", id_obis, self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        meter_id = id_obj.getValues()[1]
        if not isinstance(meter_id, str) or len(meter_id) == 0:
            LOGGER.debug("Invalid ID for OBIS code %s. Using fallback ID %s.", id_obis, self._fallback_id)
            self._trigger_id_detect_counter()
            return self._fallback_id

        self._id = meter_id
        LOGGER.debug("ID %s found with OBIS code %s.", meter_id, id_obis)
        return meter_id

    def retrieve_timestamp(self, dlms_objects: Dict[str, Any]) -> datetime:
        clock_obj = dlms_objects.get(OBIS_DEFAULT_CLOCK, None)
        if clock_obj and isinstance(clock_obj, GXDLMSClock):
            timestamp = self._extract_datetime(clock_obj)
            if timestamp:
                return timestamp

            LOGGER.warning("Unable to parse timestamp from %s. Using system time.", OBIS_DEFAULT_CLOCK)
        else:
            LOGGER.debug("No clock object found at %s. Using system time.", OBIS_DEFAULT_CLOCK)

        return datetime.utcnow()

    def get_register(self, obis: str) -> Optional[RegisterCosem]:
        return self._register_obis.get(obis, None)

    def _trigger_id_detect_counter(self):
        self._id_detect_countdown -= 1
        if self._id_detect_countdown <= 0:
            LOGGER.warning("Unable to retrieve ID after %i attempts. Using fallback ID %s from now on.",
                           COSEM_OBJECT_DETECT_ATTEMPTS, self._fallback_id)
            self._id = self._fallback_id

    @staticmethod
    def _find_id_obis(dlms_objects: Dict[str, Any]) -> Optional[str]:
        for default_obis in OBIS_DEFAULT_IDS:
            if default_obis in dlms_objects:
                return default_obis
        return None

    @staticmethod
    def _extract_datetime(clock_object: GXDLMSClock) -> Optional[datetime]:
        time_obj: GXDateTime = clock_object.getValues()[1]
        if isinstance(time_obj, GXDateTime):
            return time_obj.value
        return None



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
__version__ = "1.0.2"



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from abc import ABC, abstractmethod

from ..smartmeter.meter_data import MeterDataPoint


class DataSink(ABC):
    @abstractmethod
    async def start(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def stop(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def send(self, data_point: MeterDataPoint) -> None:
        raise NotImplementedError()



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from dataclasses import dataclass
from typing import Callable

import aioserial

from .reader import Reader, ReaderError


@dataclass
class SerialConfig:
    port: str
    baudrate: int = 9600
    data_bits: int = aioserial.EIGHTBITS
    parity: str = aioserial.PARITY_NONE
    stop_bits: int = aioserial.STOPBITS_ONE
    termination: bytes = aioserial.LF


# pylint: disable=too-few-public-methods
class SerialReader(Reader):
    def __init__(self, serial_config: SerialConfig, callback: Callable[[bytes], None]) -> None:
        super().__init__(callback)
        self._termination = serial_config.termination
        try:
            self._serial = aioserial.AioSerial(
                port=serial_config.port,
                baudrate=serial_config.baudrate,
                bytesize=serial_config.data_bits,
                parity=serial_config.parity,
                stopbits=serial_config.stop_bits
            )
        except aioserial.SerialException as ex:
            raise ReaderError(ex) from ex

    async def start_and_listen(self) -> None:
        while True:
            data: bytes = await self._serial.read_until_async(self._termination, None)
            self._callback(data)



#!/usr/bin/env python3
#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
from os import path

from setuptools import find_packages, setup

from smartmeter_datacollector.__version__ import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="smartmeter-datacollector",
    version=__version__,
    description="Smart Meter Data Collector",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scs/smartmeter-datacollector",
    project_urls={
        "Source": "https://github.com/scs/smartmeter-datacollector",
        "Bug Reports": "https://github.com/scs/smartmeter-datacollector/issues",
        "Pull Requests": "https://github.com/scs/smartmeter-datacollector/pulls",
        "SCS": "https://www.scs.ch",
    },
    author="Supercomputing Systems AG",
    author_email="info@scs.ch",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Typing :: Typed",
    ],
    license="GPLv2",
    python_requires=">=3.7",
    packages=find_packages(
        exclude=["contrib", "doc", "LICENSES", "scripts", "tests", "tests."]
    ),
    include_package_data=True,
    install_requires=[
        "aioserial==1.3.0",
        "asyncio-mqtt==0.10.0",
        "gurux-dlms==1.0.107",
        "paho-mqtt==1.5.1",
        "pyserial==3.5",
    ],
    scripts=["bin/smartmeter-datacollector"],
    zip_safe=True,
    dependency_links=[],
)



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import logging
from typing import Any, Dict, List, Optional, Tuple

from gurux_dlms import GXByteBuffer, GXDLMSClient, GXReplyData
from gurux_dlms.enums import InterfaceType, ObjectType, Security
from gurux_dlms.objects import GXDLMSData, GXDLMSObject, GXDLMSRegister
from gurux_dlms.secure import GXDLMSSecureClient

from .cosem import Cosem
from .meter_data import MeterDataPoint

LOGGER = logging.getLogger("smartmeter")


class HdlcDlmsParser:
    HDLC_BUFFER_MAX_SIZE = 5000

    def __init__(self, cosem_config: Cosem, block_cipher_key: Optional[str] = None) -> None:
        if block_cipher_key:
            self._client = GXDLMSSecureClient(
                useLogicalNameReferencing=True,
                interfaceType=InterfaceType.HDLC)
            self._client.ciphering.security = Security.ENCRYPTION
            self._client.ciphering.blockCipherKey = GXByteBuffer.hexToBytes(block_cipher_key)
        else:
            self._client = GXDLMSClient(
                useLogicalNameReferencing=True,
                interfaceType=InterfaceType.HDLC)

        self._hdlc_buffer = GXByteBuffer()
        self._dlms_data = GXReplyData()
        self._cosem = cosem_config

    def append_to_hdlc_buffer(self, data: bytes) -> None:
        if self._hdlc_buffer.getSize() > self.HDLC_BUFFER_MAX_SIZE:
            LOGGER.warning("HDLC byte-buffer > %i. Buffer is cleared, some data is lost.",
                           self.HDLC_BUFFER_MAX_SIZE)
            self._hdlc_buffer.clear()
            self._dlms_data.clear()
        self._hdlc_buffer.set(data)

    def clear_hdlc_buffer(self) -> None:
        self._hdlc_buffer.clear()

    def extract_data_from_hdlc_frames(self) -> bool:
        """
        Try to extract data fragments from HDLC frame-buffer and store it into DLMS buffer.
        HDLC buffer is being cleared.
        Returns: True if data is complete for parsing.
        """
        tmp = GXReplyData()
        try:
            LOGGER.debug("HDLC Buffer: %s", GXByteBuffer.hex(self._hdlc_buffer))
            self._client.getData(self._hdlc_buffer, tmp, self._dlms_data)
        except (ValueError, TypeError) as ex:
            LOGGER.warning("Failed to extract data from HDLC frame: '%s' Some data got lost.", ex)
            self._hdlc_buffer.clear()
            self._dlms_data.clear()
            return False

        if not self._dlms_data.isComplete():
            LOGGER.debug("HDLC frame incomplete and will not be parsed yet.")
            return False

        if self._dlms_data.isMoreData():
            LOGGER.debug("More DLMS data expected. Not yet ready to be parsed.")
            return False

        LOGGER.debug("DLMS packet complete and ready for parsing.")
        self._hdlc_buffer.clear()
        return True

    def parse_to_dlms_objects(self) -> Dict[str, GXDLMSObject]:
        parsed_objects: List[Tuple[GXDLMSObject, int]] = []
        if isinstance(self._dlms_data.value, list):
            #pylint: disable=unsubscriptable-object
            parsed_objects = self._client.parsePushObjects(self._dlms_data.value[0])
            for index, (obj, attr_ind) in enumerate(parsed_objects):
                if index == 0:
                    # Skip first (meta-data) object
                    continue
                self._client.updateValue(obj, attr_ind, self._dlms_data.value[index])
                LOGGER.debug("%s %s %s: %s", obj.objectType, obj.logicalName, attr_ind, obj.getValues()[attr_ind - 1])
        self._dlms_data.clear()
        return {obj.getName(): obj for obj, _ in parsed_objects}

    def convert_dlms_bundle_to_reader_data(self, dlms_objects: Dict[str, GXDLMSObject]) -> List[MeterDataPoint]:
        meter_id = self._cosem.retrieve_id(dlms_objects)
        timestamp = self._cosem.retrieve_timestamp(dlms_objects)

        # Extract register data
        data_points: List[MeterDataPoint] = []
        for obis, obj in filter(lambda o: o[1].getObjectType() == ObjectType.REGISTER, dlms_objects.items()):
            reg_type = self._cosem.get_register(obis)
            if reg_type and isinstance(obj, GXDLMSRegister):
                raw_value = self._extract_register_value(obj)
                if raw_value is None:
                    LOGGER.warning("No value received for %s.", obis)
                    continue
                data_point_type = reg_type.data_point_type
                try:
                    value = float(raw_value) * reg_type.scaling
                except (TypeError, ValueError, OverflowError):
                    LOGGER.warning("Invalid register value '%s'. Skipping register.", str(raw_value))
                    continue
                data_points.append(MeterDataPoint(data_point_type, value, meter_id, timestamp))
        return data_points

    @staticmethod
    def _extract_value_from_data_object(data_object: GXDLMSData) -> Optional[Any]:
        return data_object.getValues()[1]

    @staticmethod
    def _extract_register_value(register: GXDLMSRegister) -> Optional[Any]:
        return register.getValues()[1]



#
# Copyright (C) 2022 Supercomputing Systems AG
# This file is part of smartmeter-datacollector.
#
# SPDX-License-Identifier: GPL-2.0-only
# See LICENSES/README.md for more information.
#
import dataclasses
import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum


@dataclass
class MeterDataPointType:
    identifier: str
    name: str
    unit: str


class MeterDataPointTypes(Enum):
    ACTIVE_POWER_P = MeterDataPointType("ACTIVE_POWER_P", "Active Power +", "W")
    ACTIVE_POWER_P_L1 = MeterDataPointType("ACTIVE_POWER_P_L1", "Active Power L1 +", "W")
    ACTIVE_POWER_P_L2 = MeterDataPointType("ACTIVE_POWER_P_L2", "Active Power L2 +", "W")
    ACTIVE_POWER_P_L3 = MeterDataPointType("ACTIVE_POWER_P_L3", "Active Power L3 +", "W")

    ACTIVE_POWER_N = MeterDataPointType("ACTIVE_POWER_N", "Active Power -", "W")
    ACTIVE_POWER_N_L1 = MeterDataPointType("ACTIVE_POWER_N_L1", "Active Power L1 -", "W")
    ACTIVE_POWER_N_L2 = MeterDataPointType("ACTIVE_POWER_N_L2", "Active Power L2 -", "W")
    ACTIVE_POWER_N_L3 = MeterDataPointType("ACTIVE_POWER_N_L3", "Active Power L3 -", "W")

    REACTIVE_POWER_P = MeterDataPointType("REACTIVE_POWER_P", "Reactive Power +", "VA")
    REACTIVE_POWER_P_L1 = MeterDataPointType("REACTIVE_POWER_P_L1", "Reactive Power L1 +", "VA")
    REACTIVE_POWER_P_L2 = MeterDataPointType("REACTIVE_POWER_P_L2", "Reactive Power L2 +", "VA")
    REACTIVE_POWER_P_L3 = MeterDataPointType("REACTIVE_POWER_P_L3", "Reactive Power L3 +", "VA")

    REACTIVE_POWER_N = MeterDataPointType("REACTIVE_POWER_N", "Reactive Power -", "VA")
    REACTIVE_POWER_N_L1 = MeterDataPointType("REACTIVE_POWER_N_L1", "Reactive Power L1 -", "VA")
    REACTIVE_POWER_N_L2 = MeterDataPointType("REACTIVE_POWER_N_L2", "Reactive Power L2 -", "VA")
    REACTIVE_POWER_N_L3 = MeterDataPointType("REACTIVE_POWER_N_L3", "Reactive Power L3 -", "VA")

    CURRENT_L1 = MeterDataPointType("CURRENT_L1", "Current L1", "A")
    CURRENT_L2 = MeterDataPointType("CURRENT_L2", "Current L2", "A")
    CURRENT_L3 = MeterDataPointType("CURRENT_L3", "Current L3", "A")

    VOLTAGE_L1 = MeterDataPointType("VOLTAGE_L1", "Voltage L1", "V")
    VOLTAGE_L2 = MeterDataPointType("VOLTAGE_L2", "Voltage L2", "V")
    VOLTAGE_L3 = MeterDataPointType("VOLTAGE_L3", "Voltage L3", "V")

    ANGLE_UI_L1 = MeterDataPointType("ANGLE_UI_L1", "Angle U-I L1", "rad")
    ANGLE_UI_L2 = MeterDataPointType("ANGLE_UI_L2", "Angle U-I L2", "rad")
    ANGLE_UI_L3 = MeterDataPointType("ANGLE_UI_L3", "Angle U-I L3", "rad")

    ACTIVE_ENERGY_P = MeterDataPointType("ACTIVE_ENERGY_P", "Active Energy +", "Wh")
    ACTIVE_ENERGY_P_T1 = MeterDataPointType("ACTIVE_ENERGY_P_T1", "Active Energy + Tariff 1", "Wh")
    ACTIVE_ENERGY_P_T2 = MeterDataPointType("ACTIVE_ENERGY_P_T2", "Active Energy + Tariff 2", "Wh")
    ACTIVE_ENERGY_N = MeterDataPointType("ACTIVE_ENERGY_N", "Active Energy -", "Wh")
    ACTIVE_ENERGY_N_T1 = MeterDataPointType("ACTIVE_ENERGY_N", "Active Energy - Tariff 1", "Wh")
    ACTIVE_ENERGY_N_T2 = MeterDataPointType("ACTIVE_ENERGY_N", "Active Energy - Tariff 2", "Wh")

    REACTIVE_ENERGY_P = MeterDataPointType("REACTIVE_ENERGY_P", "Reactive Energy +", "VAh")
    REACTIVE_ENERGY_N = MeterDataPointType("REACTIVE_ENERGY_N", "Reactive Energy -", "VAh")

    REACTIVE_ENERGY_Q1 = MeterDataPointType("REACTIVE_ENERGY_Q1", "Reactive Energy +Ri Q1", "VAh")
    REACTIVE_ENERGY_Q2 = MeterDataPointType("REACTIVE_ENERGY_Q2", "Reactive Energy +Rc Q2", "VAh")
    REACTIVE_ENERGY_Q3 = MeterDataPointType("REACTIVE_ENERGY_Q3", "Reactive Energy -Ri Q3", "VAh")
    REACTIVE_ENERGY_Q4 = MeterDataPointType("REACTIVE_ENERGY_Q4", "Reactive Energy -Rc Q4", "VAh")

    POWER_FACTOR = MeterDataPointType("POWER_FACTOR", "Power Factor", "")
    NET_FREQUENCY = MeterDataPointType("NET_FREQUENCY", "Net Frequency any Phase", "Hz")


@dataclass
class MeterDataPoint:
    type: MeterDataPointType
    value: float
    source: str
    timestamp: datetime

    def __str__(self) -> str:
        return f"{self.source} - {self.timestamp.isoformat()} - {self.type.name}: {self.value} {self.type.unit}"

    def to_json(self) -> str:
        dict_repr = dataclasses.asdict(self)
        dict_repr['timestamp'] = self.timestamp.isoformat()
        return json.dumps(dict_repr)
